@@ -15,20 +15,18 @@
 # limitations under the License.
 
 import logging        as mod_logging
-import os             as mod_os
-import urllib         as mod_urllib
+import urllib.request as mod_urllib
 import re             as mod_re
-import pickle         as mod_pickle
-import os.path        as mod_path
+from typing import *
 
-def retrieve_all_files_urls(url):
+def retrieve_all_files_urls(url: str) -> Dict[str, str]:
     mod_logging.info('Retrieving {0}'.format(url))
     url_stream = mod_urllib.urlopen(url)
     contents = url_stream.read()
     url_stream.close()
 
     url_candidates = mod_re.findall('href="(.*?)"', contents)
-    urls = {}
+    urls: Dict[str, str] = {}
 
     for url_candidate in url_candidates:
         if url_candidate.endswith('/') and not url_candidate in url:
@@ -38,13 +36,13 @@ def retrieve_all_files_urls(url):
 
     return urls
 
-def get_files(url):
+def get_files(url: str) -> Dict[str, str]:
     mod_logging.info('Retrieving {0}'.format(url))
     url_stream = mod_urllib.urlopen(url)
     contents = url_stream.read()
     url_stream.close()
 
-    result = {}
+    result: Dict[str, str] = {}
 
     url_candidates = mod_re.findall('href="(.*?)"', contents)
     for url_candidate in url_candidates:
@@ -55,10 +53,3 @@ def get_files(url):
     mod_logging.info('Found {0} files'.format(len(result)))
 
     return result
-
-if __name__ == '__main__':
-    latitude = 45.
-    longitude = 45.
-
-    print(get_geo_elevation_data())
-
